@@ -56,12 +56,74 @@ function mount_directory()
 # Unmount - but only if mounted
 function unmount_directory()
 {
-	if mountpoint -q "$1"; then
+	#if mountpoint -q "$1"; then
+	if mountpoint "$1"; then
 		echo "Unmounting $1..."
 		umount -l "$1" || failure "Cannot unmount $1"
 	fi
 }
 
+# Check, whether union mounts are possible
+function check_union_mounts()
+{
+	[ -x /sbin/mount.aufs -o -x /usr/bin/unionfs-fuse ]
+}
+
+# union_mount -- mount a r/o file system r/w
+#	Parameters: src_file dest
+function union_mount()
+{
+	# Mount the readonly volume
+	[ ! -d "$2-mount" ] && mkdir -p "$2-mount"
+	if mount -o loop -r "$1" "$2-mount"; then
+		: ok, mount succeeded
+	else
+		rmdir "$2-mount" 2>/dev/null
+		return 1
+	fi
+
+	# Create cache and r/w directory
+	[ ! -d "$2-cache" ] && mkdir -p "$2-cache"
+	[ ! -d "$2" ] && mkdir -p "$2"
+
+	# mount as union to target
+	if [ -x /sbin/mount.aufs ]; then
+		mount -t aufs -o br:$2-cache:$2-mount none "$2"
+	elif [ -x /usr/bin/unionfs-fuse ]; then
+		unionfs-fuse -o cow,max_files=32768,hide_meta_files \
+			-o allow_other,suid,dev \
+			"$2-cache"=RW:"$2-mount"=RO "$2"
+	else
+		echo "Cannot use union_mounts!" >&2
+		false
+	fi
+	status=$?
+	if [ $status -ne 0 ]; then
+		umount "$2-mount" >/dev/null 2>&1
+		rmdir "$2" "$2-cache" "$2-mount" >/dev/null 2>&1
+	fi
+	return $status
+}
+
+# union_umount -- unmount a union_mount
+#	Paramters: mountdir
+function union_umount()
+{
+	sync
+	# Kill processes possibly still using the mount
+	for pid in `lsof 2>/dev/null | grep "$1" | grep -v unionfs | awk '{print $2}' | sort -u`
+	do
+		kill $pid
+		sleep 2		# Give some time to terminate...
+	done
+	sync
+	umount "$1"
+	rmdir "$1" >/dev/null 2>&1
+	rmdir "$1-cache" >/dev/null 2>&1
+	umount "$1-mount"
+	rmdir "$1-mount" >/dev/null 2>&1
+}
+
 # Create/Mount all filesystems for chroot environment
 function mount_pseudofilesystems()
 {
@@ -86,12 +148,23 @@ function mount_pseudofilesystems()
 	mount_directory /proc "$REMASTER_DIR/proc"
 	mount_directory /sys "$REMASTER_DIR/sys"
 	mount_directory /dev/pts "$REMASTER_DIR/dev/pts"
-	mount_directory /var/run "$REMASTER_DIR/var/run"
 	mount_directory /tmp "$REMASTER_DIR/tmp"
 	mount_directory "$REMASTER_HOME/remaster-root-home" "$REMASTER_DIR/root"
 	mount_directory "$REMASTER_HOME/remaster-apt-cache" "$REMASTER_DIR/var/cache/apt"
+	
+	if [ -d "/run" ]; then
+		HOST_VAR_RUN="/run"
+	else
+		HOST_VAR_RUN="/var/run"
+	fi
+	if [ -d "$REMASTER_DIR/run" ]; then
+		GUEST_VAR_RUN="$REMASTER_DIR/run"
+	else
+		GUEST_VAR_RUN="$REMASTER_DIR/var/run"
+	fi
+	mount_directory "$HOST_VAR_RUN" "$GUEST_VAR_RUN"
 
-	# Mount customization scripts, iff any
+	# Mount customization scripts, if any
 	if [ -e "$REMASTER_HOME/customization-scripts" ]; then
 		if [ ! -d "$REMASTER_DIR/tmp/customization-scripts" ]; then
 			mkdir "$REMASTER_DIR/tmp/customization-scripts" ||
@@ -107,15 +180,9 @@ function mount_pseudofilesystems()
 function unmount_pseudofilesystems()
 {
 	if [ -n "$REMASTER_DIR" ]; then
-		unmount_directory "$REMASTER_DIR/lib/modules/*/volatile"
-		unmount_directory "$REMASTER_DIR/tmp/customization-scripts"
-		unmount_directory "$REMASTER_DIR/var/cache/apt"
-		unmount_directory "$REMASTER_DIR/root"
-		unmount_directory "$REMASTER_DIR/tmp"
-		unmount_directory "$REMASTER_DIR/var/run"
-		unmount_directory "$REMASTER_DIR/dev/pts"
-		unmount_directory "$REMASTER_DIR/sys"
-		unmount_directory "$REMASTER_DIR/proc"
+		for i in `mount | grep " $REMASTER_DIR/" | cut -d " " -f3 | sort -r`; do
+			unmount_directory "$i"
+		done
 	fi
 }
 
@@ -240,11 +307,17 @@ function unpack_iso()
 {
 	echo "Unpacking ISO image..."
 	cp -a "$ISO_MOUNT_DIR" "$ISO_REMASTER_DIR" || failure "Failed to unpack ISO from $ISO_MOUNT_DIR to $ISO_REMASTER_DIR"
-	
+	manifest_diff
+}
+
+function manifest_diff()
+{
 	#can't trap errors with diff because of its return codes,
 	#we pass the diff's output to cut cause we strip the version number
 	if [ -e "$ISO_REMASTER_DIR/casper/filesystem.manifest" ]; then
-		diff --unchanged-group-format='' "$ISO_REMASTER_DIR/casper/filesystem.manifest" "$ISO_REMASTER_DIR/casper/filesystem.manifest-desktop" | cut -d ' ' -f 1 > "$ISO_REMASTER_DIR/casper/manifest.diff"
+		if [ -e "$ISO_REMASTER_DIR/casper/filesystem.manifest-desktop" ]; then
+			diff --unchanged-group-format='' "$ISO_REMASTER_DIR/casper/filesystem.manifest" "$ISO_REMASTER_DIR/casper/filesystem.manifest-desktop" | cut -d ' ' -f 1 > "$ISO_REMASTER_DIR/casper/manifest.diff"
+		fi
 	fi
 }
 
@@ -284,12 +357,20 @@ function prepare_rootfs_for_chroot()
 	echo "Copying resolv.conf..."
 	cp -f /etc/resolv.conf "$REMASTER_DIR/etc/resolv.conf" ||
 		failure "Failed to copy resolv.conf, error=$?"
+		
+	echo "Copying fstab/mtab..."
+	if [ -f "$REMASTER_DIR/etc/fstab" ] ; then
+		mv "$REMASTER_DIR/etc/fstab" "$REMASTER_DIR/etc/fstab.uck" ||
+			failure "Failed to copy fstab, error=$?"
+	fi
+	cp -f /etc/fstab "$REMASTER_DIR/etc/fstab" ||
+		failure "Failed to copy fstab, error=$?"
+	cp -f /etc/mtab "$REMASTER_DIR/etc/mtab" ||
+		failure "Failed to copy mtab, error=$?"
 
 	echo "Creating DBUS uuid..."
 	chroot "$REMASTER_DIR" dbus-uuidgen --ensure 1>/dev/null 2>&1
 
-	# The next is not needed by uckFlow (uses gksudo), but left in here for
-	# backwards compatibility
 	if [ -e "$REMASTER_HOME/customization-scripts/Xcookie" ] ; then
 		UCK_USER_HOME_DIR=`xauth info|grep 'Authority file'| sed "s/[ \t]//g" | sed "s/\/\.Xauthority//" | cut -d ':' -f2`
 		if [ `echo $UCK_USER_HOME_DIR | cut -d '/' -f2` == 'home' ] ; then
@@ -300,11 +381,43 @@ function prepare_rootfs_for_chroot()
 			cat "$REMASTER_HOME/customization-scripts/Xcookie" | chroot "$REMASTER_DIR" xauth merge - || failure "Failed to merge X authorization file in user directory, error=$?"
 		fi
 	fi
+	
+	echo "Deactivating initctl..."
+	chroot "$REMASTER_DIR" mv /sbin/initctl /sbin/initctl.uck_blocked
+	chroot "$REMASTER_DIR" ln -s /bin/true /sbin/initctl
+	
+	echo "Deactivating update-grub..."
+	chroot "$REMASTER_DIR" mv /usr/sbin/update-grub /usr/sbin/update-grub.uck_blocked
+	chroot "$REMASTER_DIR" ln -s /bin/true /usr/sbin/update-grub
+
+	echo "Remembering kernel update state..."
+	update_flags="reboot-required reboot-required.pkgs do-not-hibernate"
+	varrun="$REMASTER_DIR"/var/run
+	for flag in $update_flags; do
+		[ -f "$varrun"/$flag ] &&
+			mv "$varrun"/$flag "$varrun"/$flag.uck_blocked
+	done
 }
 
 function clean_rootfs_after_chroot()
 {
-	# Not used by uckFlow, but left for backward compatibility:
+	echo "Restoring kernel update state..."
+	update_flags="reboot-required reboot-required.pkgs do-not-hibernate"
+	varrun="$REMASTER_DIR"/var/run
+	for flag in $update_flags; do
+		rm -f "$varrun"/$flag
+		[ -f "$varrun"/$flag.uck_blocked ] &&
+			mv "$varrun"/$flag.uck_blocked "$varrun"/$flag
+	done
+
+	echo "Reactivating initctl..."
+	chroot "$REMASTER_DIR" rm /sbin/initctl
+	chroot "$REMASTER_DIR" mv /sbin/initctl.uck_blocked /sbin/initctl
+	
+	echo "Reactivating update-grub..."
+	chroot "$REMASTER_DIR" rm /usr/sbin/update-grub
+	chroot "$REMASTER_DIR" mv /usr/sbin/update-grub.uck_blocked /usr/sbin/update-grub 
+	
 	UCK_USER_HOME_DIR=`xauth info|grep 'Authority file'| sed "s/[ \t]//g" | sed "s/\/\.Xauthority//" | cut -d ':' -f2`
 	if [ `echo $UCK_USER_HOME_DIR | cut -d '/' -f2` == 'home' ] ; then
 		echo "Removing /home/username directory..."
@@ -316,6 +429,13 @@ function clean_rootfs_after_chroot()
 
 	echo "Removing generated resolv.conf..."
 	chroot "$REMASTER_DIR" rm -f /etc/resolv.conf
+	
+	echo "Removing generated fstab/mtab..."
+	chroot "$REMASTER_DIR" rm -f /etc/mtab
+	chroot "$REMASTER_DIR" rm -f /etc/fstab
+	if [ -f "$REMASTER_DIR/etc/fstab.uck" ] ; then
+		mv "$REMASTER_DIR/etc/fstab.uck" "$REMASTER_DIR/etc/fstab"
+	fi
 
 	unmount_pseudofilesystems
 
@@ -360,12 +480,23 @@ function pack_rootfs()
 
 		EXTRA_OPTS=""
 
-		if [ -e "$CUSTOMIZE_DIR/rootfs.sort" ] ; then
+		if [ -e "$CUSTOMIZE_DIR/rootfs.sort" ]; then
 			#FIXME: space not allowed in $CUSTOMIZE_DIR
 			EXTRA_OPTS="-sort $CUSTOMIZE_DIR/rootfs.sort"
 		fi
+		
+		#if mksquashfs version => 4.1 and guest's kernel => 2.6.30 we can enable xz compression
+		SQUASHFS_VERSION=`dpkg-query -p squashfs-tools | grep Version | cut -d ':' -f3 | cut -d '-' -f1`
+		GUEST_KERNEL_VERSION=`ls /boot/config-* | sed 's/.*config-//' | cut -d '-' -f1 | sort -r | head -n1`
+		if [ `echo -e "${SQUASHFS_VERSION}\n4.2" | sort | head -n1` = "4.2" ]; then
+			if [ `echo -e "${GUEST_KERNEL_VERSION}\n2.6.30" | sort | head -n1` = "2.6.30" ]; then
+				echo "Squashfs>=4.1, guest kernel>=2.6.30: Enabling XZ compression for squashfs..."
+				EXTRA_OPTS="${EXTRA_OPTS} -comp xz"
+			fi
+		fi
 
-		mksquashfs "$REMASTER_DIR" "$ISO_REMASTER_DIR/casper/filesystem.squashfs" $EXTRA_OPTS || failure "Failed to create squashfs image to $ISO_REMASTER_DIR/casper/filesystem.squashfs, error=$?"
+		mksquashfs "$REMASTER_DIR" "$ISO_REMASTER_DIR/casper/filesystem.squashfs" $EXTRA_OPTS ||
+			failure "Failed to create squashfs image to $ISO_REMASTER_DIR/casper/filesystem.squashfs, error=$?"
 	else
 		echo "Remastering root directory does not exists"
 	fi
@@ -456,7 +587,7 @@ function pack_iso()
 	fi
 
 	if [ "$1" = "ppc" ]; then
-		mkisofs -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
+		genisoimage -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
 			-p "Ubuntu Customization Kit - http://uck.sf.net" \
 			-probe -map "$UCK_LIBRARIES_DIR/hfs.map" -chrp-boot -iso-level 2 \
 			-part -no-desktop -r --netatalk -hfs \
@@ -467,16 +598,17 @@ function pack_iso()
 			$MKISOFS_EXTRA_OPTIONS \
 			"$ISO_REMASTER_DIR"
 	elif [ "$1" = "x86_64" ]; then
-		mkisofs -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
+		genisoimage -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
 			-b "isolinux/isolinux.bin" -c "isolinux/boot.cat" \
 			-p "Ubuntu Customization Kit - http://uck.sf.net" \
-			-no-emul-boot -V "$LIVECD_ISO_DESCRIPTION" -r -J -l \
+			-no-emul-boot -boot-load-size 4 -boot-info-table \
+			-V "$LIVECD_ISO_DESCRIPTION" -cache-inodes -r -J -l \
 			-x "$ISO_REMASTER_DIR"/casper/manifest.diff \
 			-joliet-long \
 			$MKISOFS_EXTRA_OPTIONS \
 			"$ISO_REMASTER_DIR"
 	elif [ "$1" = "ia64" ]; then
-		mkisofs -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
+		genisoimage -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
 		-b "isolinux/isolinux.bin" -c "isolinux/boot.cat" \
 		-no-emul-boot -V "$LIVECD_ISO_DESCRIPTION" -J -r \
 		-x "$ISO_REMASTER_DIR"/casper/manifest.diff \
@@ -484,7 +616,7 @@ function pack_iso()
 		$MKISOFS_EXTRA_OPTIONS \
 		"$ISO_REMASTER_DIR"
 	else
-		mkisofs -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
+		genisoimage -o "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME" \
 			-b "isolinux/isolinux.bin" -c "isolinux/boot.cat" \
 			-p "Ubuntu Customization Kit - http://uck.sf.net" \
 			-no-emul-boot -boot-load-size 4 -boot-info-table \
@@ -499,6 +631,20 @@ function pack_iso()
 	if [ $RESULT -ne 0 ]; then
 		failure "Failed to pack ISO image, error=$RESULT"
 	fi
+	
+	if [ $HYBRID = 1 ]; then
+		if [ -e "/usr/bin/isohybrid" ] ; then
+			echo "Making your ISO hybrid..."
+			/usr/bin/isohybrid "$NEW_FILES_DIR/$NEW_ISO_FILE_NAME"
+			
+			RESULT=$?
+			if [ $RESULT -ne 0 ]; then
+				failure "Failed to pack ISO image, error=$RESULT"
+			fi
+		else
+			failure "You asked for a hybrid ISO but isohybrid command was not found"
+		fi
+	fi
 }
 
 function generate_md5_for_new_iso()
